@@ -37,7 +37,7 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 #include "touch.h"
 #include "util.h"
 
-#if defined (iPlatformMsys)
+#if defined (iPlatformMsys) || defined (iPlatformWindows)
 #   include "../win32.h"
 #endif
 #if defined (iPlatformAppleDesktop)
@@ -46,6 +46,9 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 #if defined (iPlatformAppleMobile)
 #   include "ios.h"
 #endif
+#if defined (LAGRANGE_ENABLE_X11_XLIB)
+#  include "../x11.h"
+#endif
 
 #include <the_Foundation/file.h>
 #include <the_Foundation/path.h>
@@ -73,8 +76,14 @@ iDefineTypeConstructionArgs(Window,
                             type, rect, flags)
 iDefineTypeConstructionArgs(MainWindow, (iRect rect), rect)
 
+#if defined (iPlatformPcDesktop) || defined (iPlatformTerminal)
+#   define LAGRANGE_PC_MENUS
+#endif
+
 static const iMenuItem fileMenuItems_[] = {
+#if defined (LAGRANGE_MULTIPLE_WINDOWS)
     { "${menu.newwindow}", SDLK_n, KMOD_PRIMARY, "window.new" },
+#endif
     { "${menu.newtab}", SDLK_t, KMOD_PRIMARY, "tabs.new append:1" },
     { "${menu.openlocation}", SDLK_l, KMOD_PRIMARY, "navigate.focus" },
     { "${menu.reopentab}", SDLK_t, KMOD_SECONDARY, "tabs.new reopen:1" },
@@ -85,12 +94,12 @@ static const iMenuItem fileMenuItems_[] = {
     { saveToDownloads_Label, SDLK_s, KMOD_PRIMARY, "document.save" },
     { "---" },
     { "${menu.userdata}", 0, 0, "submenu id:userdatamenu" },
-//    { "${menu.downloads}", 0, 0, "downloads.open" },
-//    { "${menu.export}", 0, 0, "export" },
-#if defined (iPlatformPcDesktop)
+#if defined (LAGRANGE_PC_MENUS)
     { "---" },
     { "${menu.preferences}", preferences_KeyShortcut, "preferences" },
+#if !defined (iPlatformTerminal)
     { "${menu.fonts}", 0, 0, "open newtab:1 switch:1 url:about:fonts" },
+#endif
 #if defined (LAGRANGE_ENABLE_WINSPARKLE)
     { "${menu.update}", 0, 0, "updater.check" },
 #endif
@@ -107,16 +116,19 @@ static const iMenuItem editMenuItems_[] = {
     { "---" },
     { "${menu.copy.pagelink}", SDLK_c, KMOD_PRIMARY | KMOD_SHIFT, "document.copylink" },
     { "---" },
-    { "${macos.menu.find}", SDLK_f, KMOD_PRIMARY, "focus.set id:find.input" },
+    { "${macos.menu.find}", SDLK_f, KMOD_PRIMARY, "focus.set id:find.input id2:filter.bookmark.input" },
     { NULL }
 };
 
 static const iMenuItem viewMenuItems_[] = {
     { "${menu.show.bookmarks}", '1', leftSidebarTab_KeyModifier, "sidebar.mode arg:0 toggle:1" },
     { "${menu.show.feeds}", '2', leftSidebarTab_KeyModifier, "sidebar.mode arg:1 toggle:1" },
-    { "${menu.show.history}", '3', leftSidebarTab_KeyModifier, "sidebar.mode arg:2 toggle:1" },
+    { "${menu.show.subscriptions}", '3', leftSidebarTab_KeyModifier, "sidebar.mode arg:2 toggle:1" },
     { "${menu.show.identities}", '4', leftSidebarTab_KeyModifier, "sidebar.mode arg:3 toggle:1" },
     { "${menu.show.outline}", '5', leftSidebarTab_KeyModifier, "sidebar.mode arg:4 toggle:1" },
+    { "${menu.show.structure}", '6', leftSidebarTab_KeyModifier, "sidebar.mode arg:5 toggle:1" },
+    { "${menu.show.opendocs}", '7', leftSidebarTab_KeyModifier, "sidebar.mode arg:6 toggle:1" },
+    { "${menu.show.history}", '8', leftSidebarTab_KeyModifier, "sidebar.mode arg:7 toggle:1" },
     { "---" },
     { "${menu.sidebar.left}", leftSidebar_KeyShortcut, "sidebar.toggle" },
     { "${menu.sidebar.right}", rightSidebar_KeyShortcut, "sidebar2.toggle" },
@@ -160,16 +172,19 @@ static const iMenuItem windowMenuItems_[] = {
     { "${menu.tab.next}", 0, 0, "tabs.next" },
     { "${menu.tab.prev}", 0, 0, "tabs.prev" },
     { "${menu.duptab}", 0, 0, "tabs.new duplicate:1" },
+#if !defined (iPlatformTerminal)
+    { "${menu.movetab.newwindow}", 0, 0, "tabs.swap newwindow:1" },
     { "---" },
     { "${menu.window.min}", 0, 0, "window.minimize" },
     { "${menu.window.max}", 0, 0, "window.maximize" },
     { "${menu.window.full}", 0, 0, "window.fullscreen" },
+#endif
     { "---" },
     { NULL }
 };
 
 static const iMenuItem helpMenuItems_[] = {
-#if defined (iPlatformPcDesktop)
+#if defined (LAGRANGE_PC_MENUS)
     { "${menu.help}", SDLK_F1, 0, "!open newtab:1 switch:1 url:about:help" },
 #else
     { "${menu.help}", 0, 0, "!open newtab:1 switch:1 url:about:help" },
@@ -178,7 +193,7 @@ static const iMenuItem helpMenuItems_[] = {
     { "---" },
     { "${menu.aboutpages}", 0, 0, "!open newtab:1 switch:1 url:about:about" },
     { "${menu.debug}", 0, 0, "!open newtab:1 switch:1 url:about:debug" },
-#if defined (iPlatformPcDesktop)
+#if defined (LAGRANGE_PC_MENUS)
     { "---" },
     { "${menu.aboutapp}", 0, 0, "!open newtab:1 switch:1 url:about:lagrange" },
 #endif
@@ -195,6 +210,10 @@ const iMenuItem topLevelMenus_Window[7] = {
     { "${menu.title.help}", 0, 0, (const void *) helpMenuItems_ },
 };
 
+size_t numWindowMenuItems_Window(void) {
+    return iElemCount(windowMenuItems_) - 1; /* don't count the terminator */
+}
+
 #if defined (LAGRANGE_MAC_MENUBAR)
 
 static iBool macMenusInserted_;
@@ -242,6 +261,17 @@ int numRoots_Window(const iWindow *d) {
     return num;
 }
 
+static void windowSizeChanged_Window_(iWindow *d) {
+    setCurrent_Window(d);
+    iRoot *root = d->roots[0];
+    iRect *rect = &root->widget->rect;
+    rect->pos   = zero_I2();
+    rect->size  = d->size;
+    setCurrent_Root(root);
+    updatePadding_Root(root);
+    arrange_Widget(root->widget);
+}
+
 static void windowSizeChanged_MainWindow_(iMainWindow *d) {
     const int numRoots = numRoots_Window(as_Window(d));
     const iInt2 rootSize = d->base.size;
@@ -333,6 +363,24 @@ static void setupUserInterface_MainWindow(iMainWindow *d) {
 #endif
 }
 
+static iBool updateSize_Window_(iWindow *d, iBool notifyAlways) {
+    iInt2 *size = &d->size;
+    const iInt2 oldSize = *size;
+    SDL_GetRendererOutputSize(d->render, &size->x, &size->y);
+    const iBool hasChanged = !isEqual_I2(oldSize, *size);
+    if (hasChanged) {
+        windowSizeChanged_Window_(d);
+        postRefresh_Window(d);
+    }
+    if (!isResizing_ && (hasChanged || notifyAlways)) {
+        iRoot *root = d->roots[0];
+        postCommandf_Root(root, "window.resized width:%d height:%d", size->x, size->y);
+        postCommand_Root(root, "widget.overflow"); /* check bounds with updated sizes */
+        postRefresh_Window(d);
+    }
+    return hasChanged;
+}
+
 static iBool updateSize_MainWindow_(iMainWindow *d, iBool notifyAlways) {
     iInt2 *size = &d->base.size;
     const iInt2 oldSize = *size;
@@ -396,6 +444,15 @@ static float pixelRatio_Window_(const iWindow *d) {
 float displayDensity_Android(void);
 #endif
 
+#if defined (iPlatformWindows)
+static void setenv(const char *name, const char *value, int overwrite) {
+    iUnused(overwrite);
+    SetEnvironmentVariableW(
+        data_Block(collect_Block(toUtf16_String(collectNewCStr_String(name)))),
+        data_Block(collect_Block(toUtf16_String(collectNewCStr_String(value)))));
+}
+#endif
+
 static float displayScale_Window_(const iWindow *d) {
 #if !defined (iPlatformTerminal)
     /* The environment variable LAGRANGE_OVERRIDE_DPI can be used to override the automatic
@@ -421,7 +478,7 @@ static float displayScale_Window_(const iWindow *d) {
     /* Apple UI sizes are fixed and only scaled by pixel ratio. */
     /* TODO: iOS text size setting? */
     return 1.0f;
-#elif defined (iPlatformMsys)
+#elif defined (iPlatformMsys) || defined (iPlatformWindows)
     iUnused(d);
     return desktopDPI_Win32();
 #elif defined (iPlatformAndroidMobile)
@@ -512,7 +569,7 @@ SDL_HitTestResult hitTest_MainWindow(const iMainWindow *d, iInt2 pos) {
 
 void create_Window_(iWindow *d, iRect rect, uint32_t flags) {
     flags |= SDL_WINDOW_ALLOW_HIGHDPI | SDL_WINDOW_HIDDEN;
-    if (d->type == main_WindowType) {
+    if (d->type == main_WindowType || d->type == extra_WindowType) {
         flags |= SDL_WINDOW_RESIZABLE;
 #if defined (LAGRANGE_ENABLE_CUSTOM_FRAME)
         if (prefs_App()->customFrame) {
@@ -562,7 +619,7 @@ void create_Window_(iWindow *d, iRect rect, uint32_t flags) {
             exit(-4);
         }
     }
-#if defined(LAGRANGE_ENABLE_CUSTOM_FRAME)
+#if defined (LAGRANGE_ENABLE_CUSTOM_FRAME)
     if (type_Window(d) == main_WindowType && prefs_App()->customFrame) {
         /* Register a handler for window hit testing (drag, resize). */
         SDL_SetWindowHitTest(d->win, hitTest_MainWindow_, d);
@@ -601,6 +658,7 @@ void init_Window(iWindow *d, enum iWindowType type, iRect rect, uint32_t flags)
     d->hover         = NULL;
     d->lastHover     = NULL;
     d->mouseGrab     = NULL;
+    d->keyPriority   = NULL;
     d->focus         = NULL;
     d->pendingCursor = NULL;
     d->isExposed     = (deviceType_App() != desktop_AppDeviceType);
@@ -622,18 +680,18 @@ void init_Window(iWindow *d, enum iWindowType type, iRect rect, uint32_t flags)
     /* Renderer info. */ {
         SDL_RendererInfo info;
         SDL_GetRendererInfo(d->render, &info);
-#if !defined (NDEBUG)
+#   if !defined (NDEBUG)
         printf("[window] renderer: %s%s\n",
                info.name,
                info.flags & SDL_RENDERER_ACCELERATED ? " (accelerated)" : "");
-#endif
+#   endif
     }
-#endif /* !iPlatformTerminal */
-#if defined (iPlatformMsys)
+#   if defined (iPlatformMsys) || defined (iPlatformWindows)
     if (type == extra_WindowType) {
         enableDarkMode_SDLWindow(d->win);
     }
-#endif
+#   endif
+#endif /* !iPlatformTerminal */
     drawBlank_Window_(d);
     d->pixelRatio   = pixelRatio_Window_(d); /* point/pixel conversion */
     d->displayScale = displayScale_Window_(d);
@@ -678,14 +736,16 @@ void deinit_Window(iWindow *d) {
 }
 
 static void setWindowIcon_Window_(iWindow *d) {
-#if defined (iPlatformMsys)
+#if !defined (iPlatformTerminal)
+#   if defined (iPlatformMsys) || defined (iPlatformWindows)
     useExecutableIconResource_SDLWindow(d->win);
-#endif
-#if defined (iPlatformLinux) && !defined (iPlatformTerminal)
+#   endif
+#   if defined (iPlatformLinux)
     SDL_Surface *surf = loadImage_(&imageLagrange64_Resources, 0);
     SDL_SetWindowIcon(d->win, surf);
     free(surf->pixels);
     SDL_FreeSurface(surf);
+#   endif
 #endif
     iUnused(d); /* other platforms */
 }
@@ -695,6 +755,9 @@ void init_MainWindow(iMainWindow *d, iRect rect) {
     theMainWindow_ = d;
     d->enableBackBuf = iFalse;
     uint32_t flags = 0;
+
+    d->place.desktop = -1; // init to unknown
+
 #if defined (iPlatformAppleDesktop)
     SDL_SetHint(SDL_HINT_RENDER_DRIVER, shouldDefaultToMetalRenderer_MacOS() ? "metal" : "opengl");
     flags |= shouldDefaultToMetalRenderer_MacOS() ? SDL_WINDOW_METAL : SDL_WINDOW_OPENGL;
@@ -720,7 +783,7 @@ void init_MainWindow(iMainWindow *d, iRect rect) {
     SDL_SetHint(SDL_HINT_RENDER_VSYNC, "1");
     init_Window(&d->base, main_WindowType, rect, flags);
     d->isDrawFrozen           = iTrue;
-#if defined (iPlatformMsys)
+#if defined (iPlatformMsys) || defined (iPlatformWindows)
     /* It is less glitchy to allow drawing as early as possible, to avoid white
        flashes when windows are created. */
     d->isDrawFrozen           = iFalse;
@@ -746,7 +809,7 @@ void init_MainWindow(iMainWindow *d, iRect rect) {
         SDL_RendererInfo info;
         SDL_GetRendererInfo(d->base.render, &info);
         isOpenGLRenderer_ = !iCmpStr(info.name, "opengl");
-#if !defined(NDEBUG) && !defined (iPlatformTerminal)
+#if !defined (NDEBUG) && !defined (iPlatformTerminal)
         printf("[window] max texture size: %d x %d\n",
                info.max_texture_width,
                info.max_texture_height);
@@ -756,16 +819,20 @@ void init_MainWindow(iMainWindow *d, iRect rect) {
         }
 #endif
     }
-#if defined (iPlatformMsys)
-    SDL_SetWindowMinimumSize(d->base.win, minSize.x * d->base.displayScale, minSize.y * d->base.displayScale);
+#if !defined (iPlatformTerminal)
+#   if defined (iPlatformMsys) || defined (iPlatformWindows)
+    SDL_SetWindowMinimumSize(
+        d->base.win, minSize.x * d->base.displayScale, minSize.y * d->base.displayScale);
     enableDarkMode_SDLWindow(d->base.win);
-#endif
-#if defined (iPlatformLinux) && !defined (iPlatformTerminal)
-    SDL_SetWindowMinimumSize(d->base.win, minSize.x * d->base.pixelRatio, minSize.y * d->base.pixelRatio);
-#endif
-#if defined (iPlatformAppleMobile)
+#   endif
+#   if defined (iPlatformLinux)
+    SDL_SetWindowMinimumSize(
+        d->base.win, minSize.x * d->base.pixelRatio, minSize.y * d->base.pixelRatio);
+#   endif
+#   if defined (iPlatformAppleMobile)
     setupWindow_iOS(as_Window(d));
-#endif
+#   endif
+#endif /* !defined (iPlatformTerminal) */
     setWindowIcon_Window_(as_Window(d));
     setCurrent_Text(d->base.text);
     SDL_GetRendererOutputSize(d->base.render, &d->base.size.x, &d->base.size.y);
@@ -783,7 +850,7 @@ void init_MainWindow(iMainWindow *d, iRect rect) {
         SDL_Surface *surf = loadImage_(&imageLogo_Resources, 0);
         d->logo = SDL_CreateTextureFromSurface(d->base.render, surf);
         SDL_SetTextureBlendMode(d->logo, SDL_BLENDMODE_BLEND);
-#if SDL_VERSION_ATLEAST(2, 0, 12)
+#if SDL_VERSION_ATLEAST(2, 0, 12) && !defined (iPlatformTerminal)
         SDL_SetTextureScaleMode(d->logo, SDL_ScaleModeBest);
 #endif
         free(surf->pixels);
@@ -850,7 +917,6 @@ iBool isFullscreen_MainWindow(const iMainWindow *d) {
 }
 
 iRoot *findRoot_Window(const iWindow *d, const iWidget *widget) {
-
     while (widget->parent) {
         widget = widget->parent;
     }
@@ -1006,6 +1072,15 @@ static iBool handleWindowEvent_Window_(iWindow *d, const SDL_WindowEvent *ev) {
                 checkPixelRatioChange_Window_(d);
             }
             return iFalse;
+        case SDL_WINDOWEVENT_RESIZED:
+            if (d->isMinimized) {
+                return iTrue;
+            }
+            closePopups_App(iFalse);
+            checkPixelRatioChange_Window_(d);
+            updateSize_Window_(d, iTrue);
+            postRefresh_Window(d);
+            return iTrue;
         case SDL_WINDOWEVENT_FOCUS_GAINED:
             if (d->type == extra_WindowType) {
                 d->focusGainedAt = SDL_GetTicks();
@@ -1062,13 +1137,19 @@ static void savePlace_MainWindow_(iAny *mainWindow) {
         SDL_GetWindowPosition(d->base.win, &newPos.x, &newPos.y);
         d->place.normalRect.pos = newPos;
         iInt2 border = zero_I2();
-#if !defined(iPlatformApple) && !defined (iPlatformTerminal)
+#if !defined (iPlatformApple) && !defined (iPlatformTerminal)
         SDL_GetWindowBordersSize(d->base.win, &border.y, &border.x, NULL, NULL);
         iAssert(~SDL_GetWindowFlags(d->base.win) & SDL_WINDOW_MAXIMIZED);
 #endif
         d->place.normalRect.pos =
             max_I2(zero_I2(), sub_I2(d->place.normalRect.pos, border));
     }
+#if defined (LAGRANGE_ENABLE_X11_XLIB)
+    unsigned long desk;
+    if (getWindowDesktop_X11(d->base.win, &desk)) {
+        d->place.desktop = (int) desk;
+    }
+#endif
 }
 
 static void notifyHovered_Window_(iWindow *d) {
@@ -1096,7 +1177,7 @@ static iBool handleWindowEvent_MainWindow_(iMainWindow *d, const SDL_WindowEvent
         return iFalse;
     }
     switch (ev->event) {
-#if defined(iPlatformDesktop)
+#if defined (iPlatformDesktop)
         case SDL_WINDOWEVENT_EXPOSED:
             d->base.isExposed = iTrue;
             /* Since we are manually controlling when to redraw the window, we are responsible
@@ -1104,7 +1185,7 @@ static iBool handleWindowEvent_MainWindow_(iMainWindow *d, const SDL_WindowEvent
                circumstances (e.g., under openbox), not doing this would mean that the window
                is missing contents until other events trigger a refresh. */
             postRefresh_Window(d);
-#if defined(LAGRANGE_ENABLE_WINDOWPOS_FIX)
+#if defined (LAGRANGE_ENABLE_WINDOWPOS_FIX)
             if (d->place.initialPos.x >= 0) {
                 /* Must not move a maximized window. */
                 if (snap_MainWindow(d) == 0) {
@@ -1130,7 +1211,7 @@ static iBool handleWindowEvent_MainWindow_(iMainWindow *d, const SDL_WindowEvent
                 d->base.isMinimized = iTrue;
                 return iFalse;
             }
-#if defined(LAGRANGE_ENABLE_CUSTOM_FRAME)
+#if defined (LAGRANGE_ENABLE_CUSTOM_FRAME)
             /* Set the snap position depending on where the mouse cursor is. */
             if (prefs_App()->customFrame) {
                 SDL_Rect usable;
@@ -1169,7 +1250,6 @@ static iBool handleWindowEvent_MainWindow_(iMainWindow *d, const SDL_WindowEvent
         }
         case SDL_WINDOWEVENT_RESIZED:
             if (d->base.isMinimized) {
-                // updateSize_Window_(d, iTrue);
                 return iTrue;
             }
             closePopups_App(iFalse);
@@ -1292,7 +1372,7 @@ iBool processEvent_Window(iWindow *d, const SDL_Event *ev) {
         case SDL_RENDER_TARGETS_RESET:
         case SDL_RENDER_DEVICE_RESET: {
             if (mw || extraw) {
-                invalidate_Window_(d, iTrue /* force full reset */);
+                invalidate_Window_(d, iFalse);
             }
             break;
         }
@@ -1308,32 +1388,6 @@ iBool processEvent_Window(iWindow *d, const SDL_Event *ev) {
                 postCommand_App("media.player.update"); /* in case a player needs updating */
                 return iFalse; /* unfreeze all frozen windows */
             }
-#if 0
-            if (event.type == SDL_USEREVENT && isCommand_UserEvent(ev, "window.sysframe") && mw) {
-                /* This command is sent on Android to update the keyboard height. */
-                const char *cmd = command_UserEvent(ev);
-                /*
-                    0
-                    |
-                 top
-                 |  |
-                 | bottom (top of keyboard)   :
-                 |  |                         : keyboardHeight
-                 maxDrawableHeight            :
-                    |
-                   fullheight
-                 */
-                const int top    = argLabel_Command(cmd, "top");
-                const int bottom = argLabel_Command(cmd, "bottom");
-                const int full   = argLabel_Command(cmd, "fullheight");
-                //if (!SDL_IsScreenKeyboardShown(mw->base.win)) {
-                if (bottom == full) {
-                    mw->maxDrawableHeight = bottom - top;
-                }
-                setKeyboardHeight_MainWindow(mw, top + mw->maxDrawableHeight - bottom);
-                return iTrue;
-            }
-#endif
             if (processEvent_Touch(&event)) {
                 return iTrue;
             }
@@ -1372,18 +1426,23 @@ iBool processEvent_Window(iWindow *d, const SDL_Event *ev) {
                     }
                 }
             }
-//            const iWidget *oldHover = d->hover;
             iBool wasUsed = iFalse;
             /* Dispatch first to the mouse-grabbed widget. */
-//            iWidget *widget = d->root.widget;
             if (event.type == SDL_MOUSEMOTION || event.type == SDL_MOUSEWHEEL ||
                 event.type == SDL_MOUSEBUTTONUP || event.type == SDL_MOUSEBUTTONDOWN) {
                 if (mouseGrab_Widget()) {
                     iWidget *grabbed = mouseGrab_Widget();
-                    setCurrent_Root(grabbed->root /* findRoot_Window(d, grabbed)*/);
+                    setCurrent_Root(grabbed->root);
                     wasUsed = dispatchEvent_Widget(grabbed, &event);
                 }
             }
+            /* If there is a priority handler for key events, offer the event to it first.
+               This is similar to mouse grabbing, but the handler can refuse the event. */
+            if (d->keyPriority && (event.type == SDL_KEYDOWN || event.type == SDL_KEYDOWN)) {
+                /* The event is processed directly by the widget only, not dispatched to
+                   the widget subtree. When dispatching, children still get priority. */
+                wasUsed = class_Widget(d->keyPriority)->processEvent(d->keyPriority, &event);
+            }
             /* Dispatch the event to the tree of widgets. */
             if (!wasUsed) {
                 wasUsed = dispatchEvent_Window(d, &event);
@@ -1435,7 +1494,7 @@ iBool processEvent_Window(iWindow *d, const SDL_Event *ev) {
                     }
                 }
             }
-            if (event.type == SDL_MOUSEMOTION) {
+            if (event.type == SDL_MOUSEMOTION && event.motion.windowID == id_Window(d)) {
                 applyCursor_Window_(d);
             }
             return wasUsed;
@@ -1449,6 +1508,7 @@ iBool setKeyRoot_Window(iWindow *d, iRoot *root) {
         d->keyRoot = root;
         postCommand_App("keyroot.changed");
         postRefresh_Window(d);
+        d->keyPriority = NULL;
         return iTrue;
     }
     return iFalse;
@@ -1693,23 +1753,22 @@ void draw_MainWindow(iMainWindow *d) {
        when the custom frame is being used. */ {
         setCurrent_Root(w->roots[0]);
         iColor back;
-        if (isMobile_Platform()) {
+#if defined (LAGRANGE_ENABLE_CUSTOM_FRAME)        
+        if (prefs_App()->customFrame && numRoots_Window(as_Window(d)) == 1) {
+            back = get_Color(gotFocus && d->place.snap != maximized_WindowSnap &&
+                                        ~winFlags & SDL_WINDOW_FULLSCREEN_DESKTOP
+                                    ? uiAnnotation_ColorId
+                                    : uiSeparator_ColorId);
+        }
+        else
+#endif
+        {
             back = get_Color(uiBackground_ColorId);
             if (deviceType_App() == phone_AppDeviceType) {
                 /* Page background extends to safe area, so fill it completely. */
                 back = get_Color(tmBackground_ColorId);
             }
         }
-        else {
-#if defined (LAGRANGE_ENABLE_CUSTOM_FRAME)
-            back = get_Color(gotFocus && d->place.snap != maximized_WindowSnap &&
-                                     ~winFlags & SDL_WINDOW_FULLSCREEN_DESKTOP
-                                 ? uiAnnotation_ColorId
-                                 : uiSeparator_ColorId);
-#else
-            back = get_Color(uiBackground_ColorId);
-#endif
-        }
         unsetClip_Paint(&p); /* update clip to full window */
         SDL_SetRenderDrawColor(w->render, back.r, back.g, back.b, 255);
         SDL_RenderClear(w->render);
@@ -1769,11 +1828,16 @@ void draw_MainWindow(iMainWindow *d) {
                     }
                     if (root == w->keyRoot) {
                         const iBool isDark = isDark_ColorTheme(colorTheme_App());
-                        fillRect_Paint(&p, (iRect){
-                            topLeft_Rect(bounds),
-                            init_I2(width_Rect(bounds), gap_UI / 2)
-                        }, isDark ? uiBackgroundSelected_ColorId
-                                  : uiIcon_ColorId);
+                        const int keyColor = isDark ? uiBackgroundSelected_ColorId
+                                                    : uiIcon_ColorId;
+                        const iInt2 keySize = init_I2(width_Rect(bounds), gap_UI / 2);
+                        fillRect_Paint(&p, (iRect){ topLeft_Rect(bounds), keySize }, keyColor);
+                        if (keySize.x < d->base.size.x) {
+                            fillRect_Paint(&p, (iRect){
+                                addY_I2(bottomLeft_Rect(bounds), -gap_UI / 2),
+                                keySize
+                            }, keyColor);
+                        }
                     }
                 }
             }
@@ -1815,7 +1879,9 @@ void resize_MainWindow(iMainWindow *d, int w, int h) {
 }
 
 void setTitle_Window(iWindow *d, const iString *title) {
-    SDL_SetWindowTitle(d->win, cstr_String(title));
+    iString *clean = collect_String(copy_String(title));
+    replaceRegExp_String(clean, iClob(makeAnsiEscapePattern_Text(iTrue)), "", NULL, NULL);
+    SDL_SetWindowTitle(d->win, cstr_String(clean));
     iLabelWidget *bar = findChild_Widget(get_Root()->widget, "winbar.title");
     if (bar) {
         updateText_LabelWidget(bar, title);
@@ -1826,7 +1892,10 @@ void setUiScale_Window(iWindow *d, float uiScale) {
     if (uiScale <= 0.0f) {
         uiScale = 1.0f;
     }
-    uiScale = iClamp(uiScale, 0.5f, 4.0f);
+    uiScale = iClamp(uiScale,
+                     0.5f,
+                     deviceType_App() == phone_AppDeviceType ? 1.5f :
+                     deviceType_App() == tablet_AppDeviceType ? 2.0f : 4.0f);
     if (d) {
         if (iAbs(d->uiScale - uiScale) > 0.0001f) {
             d->uiScale = uiScale;
@@ -2083,7 +2152,8 @@ void setSplitMode_MainWindow(iMainWindow *d, int splitFlags) {
         /* Add some room for the active root indicator. */
         for (int i = 0; i < 2; i++) {
             if (w->roots[i]) {
-                w->roots[i]->widget->padding[1] = (splitMode ? 1 : 0);
+                w->roots[i]->widget->padding[1] = (splitMode ? gap_UI / 2 : 0);
+                w->roots[i]->widget->padding[3] = (splitMode ? gap_UI / 2 : 0);
             }
         }
         d->splitMode = splitMode;
@@ -2091,7 +2161,9 @@ void setSplitMode_MainWindow(iMainWindow *d, int splitFlags) {
         postCommand_App("window.resized"); /* not really, but widgets may need to change layout */
 #if defined (LAGRANGE_ENABLE_CUSTOM_FRAME)
         /* Update custom frame controls. */ {
-            const iBool hideCtl0 = numRoots_Window(as_Window(d)) != 1;
+            const iBool isVerticalSplit = (splitFlags & vertical_WindowSplit) != 0;
+            const iBool hideCtl0 = numRoots_Window(as_Window(d)) > 1 && !isVerticalSplit;
+            const iBool hideCtl1 = numRoots_Window(as_Window(d)) > 1 && isVerticalSplit;
             iWidget *winBar = findChild_Widget(d->base.roots[0]->widget, "winbar");
             if (winBar) {
                 setFlags_Widget(
@@ -2099,13 +2171,19 @@ void setSplitMode_MainWindow(iMainWindow *d, int splitFlags) {
                 setFlags_Widget(
                     findChild_Widget(winBar, "winbar.max"), hidden_WidgetFlag, hideCtl0);
                 setFlags_Widget(
-                    findChild_Widget(winBar, "winbar.close"), hidden_WidgetFlag, hideCtl0);
+                    findChild_Widget(winBar, "winbar.close"), hidden_WidgetFlag, hideCtl0);                
                 if (d->base.roots[1]) {
                     winBar = findChild_Widget(d->base.roots[1]->widget, "winbar");
                     setFlags_Widget(
                         findChild_Widget(winBar, "winbar.icon"), hidden_WidgetFlag, iTrue);
                     setFlags_Widget(
                         findChild_Widget(winBar, "winbar.app"), hidden_WidgetFlag, iTrue);
+                    setFlags_Widget(
+                        findChild_Widget(winBar, "winbar.min"), hidden_WidgetFlag, hideCtl1);
+                    setFlags_Widget(
+                        findChild_Widget(winBar, "winbar.max"), hidden_WidgetFlag, hideCtl1);
+                    setFlags_Widget(
+                        findChild_Widget(winBar, "winbar.close"), hidden_WidgetFlag, hideCtl1);                
                 }
             }
         }
@@ -2233,7 +2311,7 @@ iWindow *newPopup_Window(iInt2 screenPos, iWidget *rootWidget) {
     setForceSoftwareRender_App(iTrue);
 #endif
     SDL_Rect usableRect;
-    SDL_GetDisplayUsableBounds(SDL_GetWindowDisplayIndex(get_MainWindow()->base.win),
+    SDL_GetDisplayUsableBounds(SDL_GetWindowDisplayIndex(get_Window()->win),
                                &usableRect);
     const float pixelRatio = get_Window()->pixelRatio;
     iRect winRect = (iRect){ screenPos,
@@ -2268,7 +2346,7 @@ iWindow *newPopup_Window(iInt2 screenPos, iWidget *rootWidget) {
     setRoot_Widget(rootWidget, root);
     setDrawBufferEnabled_Widget(rootWidget, iFalse);
     setForceSoftwareRender_App(oldSw);
-#if !defined(NDEBUG)
+#if !defined (NDEBUG)
     stop_PerfTimer(newPopup_Window);
 #endif
     return win;
@@ -2288,6 +2366,7 @@ iWindow *newExtra_Window(iWidget *rootWidget) {
     setCurrent_Window(win);
     iWidget *frameRoot = new_Widget();
     setFlags_Widget(frameRoot, arrangeSize_WidgetFlag | focusRoot_WidgetFlag, iTrue);
+    setFrameColor_Widget(rootWidget, none_ColorId);
     setCommandHandler_Widget(frameRoot, handleRootCommands_Widget);
     setRoot_Widget(rootWidget, root);
     addChild_Widget(frameRoot, rootWidget);
@@ -2370,6 +2449,11 @@ const iArray *updateBookmarksMenu_Widget(iWidget *menu) {
             }
 #endif
         }
+        iString *setIdentArg = NULL;
+        if (!isEmpty_String(&bm->identity)) {
+            setIdentArg = copy_String(&bm->identity);
+            prependCStr_String(setIdentArg, " setident:");
+        }
         pushBack_Array(
             dest,
             &(iMenuItem){ format_CStr("%s %s", cstr_String(&iconStr), cstr_String(title)),
@@ -2377,7 +2461,10 @@ const iArray *updateBookmarksMenu_Widget(iWidget *menu) {
                           0,
                           isFolder_Bookmark(bm)
                               ? format_CStr("submenu id:bfmenu.%d", id_Bookmark(bm))
-                              : format_CStr("!open url:%s", cstr_String(&bm->url)) });
+                              : format_CStr("!open%s url:%s",
+                                            setIdentArg ? cstr_String(setIdentArg) : "",
+                                            cstr_String(&bm->url)) });
+        delete_String(setIdentArg);
         deinit_String(&iconStr);
     }
     /* Create folder menus. */
